from kiwoom.config.const import EVENTS


def valid_event(event):
    """
    Returns True if event is valid, else False.

    :param event: str
        One of the pre-defined event names in string.
    :return: bool
    """
    if event not in EVENTS:
        print(f"{event} is not a valid event.\nSelect one of {EVENTS}.")
        return False
    return True



from collections import defaultdict


TYPES = ('single', 'multi')


class Share:
    """
    Share Class handles common data shared between Signals and Slots

    There are 4 types of data.
    1. args
    2. single
    3. multi
    4. history

    Each of data has 3 methods.
    1. get
    2. remove
    3. update/extend

    Dictionary-like 4 methods in common for single and multi data.
    1. keys
    2. has_key
    3. get
    4. items
    """
    def __init__(self):
        self.args = dict()
        self.single = defaultdict(dict)
        self.multi = defaultdict(lambda: defaultdict(list))
        self.history = defaultdict(lambda: defaultdict(list))

    """
    Dictionary-like Methods
    """
    def keys(self, typ, fn=None):
        """
        Returns keys stored for given fn.         
        
        :param typ: str
            must be one of ('single', 'multi')
        :param fn: str or None
            method name used when calling one of updating methods
            if fn is None, returns all 'fn' stored 
        """
        if typ.lower() not in TYPES:
            raise KeyError(f"Given type must be one of {TYPES}, not '{typ}'.")
        
        dic = getattr(self, typ)
        if fn is None:
            return dic.keys()
        if fn in dic:
            return dic[fn].keys()
        raise KeyError(f"Given fn, '{fn}', is not in {typ} data.")

    def has_key(self, typ, fn, key=None):
        """
        Returns if key stored in fn of given typ. 
        
        :param typ: str
            must be one of ('single', 'multi')
        :param fn: str
            method name used when calling one of updating methods
        :param key: str or None
            key that needs to be checked whether or not it is stored
            if key is None, check whether fn is stored in share or not
        """
        if key is None:
            return fn in self.keys(typ)
        return key in self.keys(typ, fn)

    def get(self, typ, fn, key=None):
        """
        Returns if key stored in fn of given typ.
        
        :param typ: str
            must be one of ('single', 'multi')
        :param fn: str
            method name used when calling one of updating methods
        :param key: str or None
            key that needs to be checked whether or not it is stored
            if key is None, get whole data stored for fn
        """
        if self.has_key(typ, fn, key):
            dic = getattr(self, typ)
            if key is None:
                return dic[fn]
            return dic[fn][key]
        
        # Raise an error
        if self.has_key(typ, fn):
            raise KeyError(f"Given key, '{key}', is not in {typ} data.")
        raise KeyError(f"Given fn, '{fn}', is not in {typ} data.")

    def items(self, typ, fn):
        """
        Returns a dict_items with (key, val) pairs for given fn. 

        :param typ: str
            must be one of ('single', 'multi')
        :param fn: str
            method name used when calling one of updating methods
        """
        dic = self.get(typ, fn)
        return dic.items()

    """
    Arg Data
    """
    def isin_args(self, fn, key=None):
        if key is None:
            return fn in self.args
        return key in self.args[fn]

    def get_args(self, fn, key=None):
        if key is None:
            return self.args[fn]
        return self.args[fn][key]

    def update_args(self, fn, args):
        self.args[fn] = args

    def remove_args(self, fn):
        if fn in self.args:
            del self.args[fn]

    """
    Single Data
    """
    def isin_single(self, fn, key=None):
        return self.has_key('single', fn, key)

    def get_single(self, fn, key=None):
        return self.get('single', fn, key)

    def update_single(self, fn, key, val):
        self.single[fn][key] = val

    def remove_single(self, fn, key=None):
        if key is None:
            if fn in self.single:
                del self.single[fn]
            return
        if fn in self.single:
            if key in self.single[fn]:
                del self.single[fn][key]

    """
    Multi Data
    """
    def isin_multi(self, fn, key=None):
        return self.has_key('multi', fn, key)

    def get_multi(self, fn, key=None):
        return self.get('multi', fn, key)

    def extend_multi(self, fn, key, vals):
        # To Extend list to original data
        self.multi[fn][key] += vals

    def remove_multi(self, fn):
        if fn in self.multi:
            del self.multi[fn]

    """
    History Data
    """
    def get_history(self, code=None, key=None):
        if code is None:
            return self.history
        elif key is None:
            return self.history[code]
        return self.history[code][key]

    def extend_history(self, code, key, vals):
        # To Extend list to original data
        self.history[code][key] += vals

    def remove_history(self, code):
        if code in self.history:
            del self.history[code]



import sys
from functools import wraps
from traceback import format_exc

from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import QApplication

from kiwoom.utils.general import clock


def timer(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        # Extract bot instance
        bot = args[0]

        def unloop():
            if not bot.connected():
                print(f'[{clock()}] Timeover for login. Please try again.')
                # Close all pop-up windows from Kiwoom
                app = QApplication.instance()
                app.closeAllWindows()
                bot.api.unloop()

        # Set timer waiting for 1 minute
        timer = QTimer(bot.api)
        timer.singleShot(60 * 1000, unloop)

        # Execute signal function
        fn(*args, **kwargs)

    return wrapper


class Downloader:
    # Variable to trace download status
    nrq = 0

    # Decorates Signal.histories()
    @staticmethod
    def watcher(fn):
        # To keep docstring of fn
        @wraps(fn)
        # Define a wrapper
        def wrapper(*args, **kwargs):
            # Extract bot instance
            bot = args[0]
            Downloader.nrq = 0

            # If not connected, exit program.
            if not bot.connected():
                return sys.exit(0)

            # Slot function to be called AFTER 10 min
            def watcher():
                # Number of request to server
                nrq = bot.share.get_single('histories', 'nrq')

                # If disconnected or froze by some issue, exit program
                if not bot.connected() or nrq == Downloader.nrq:
                    # Close all pop-up windows from Kiwoom
                    app = QApplication.instance()
                    app.closeAllWindows()

                    # Set restart param and unloop
                    bot.share.update_single('history', 'restart', True)
                    bot.api.unloop()

                    # Exit script after 60 seconds
                    print(f'\n[{clock()}] As downloader has frozen, exiting the program in 60 sec.')
                    return QTimer(app).singleShot(60, sys.exit)

                # Update nrq
                Downloader.nrq = nrq

            # Set timer for every 10 minutes
            timer = QTimer(bot.api)
            timer.start(10 * 60 * 1000)
            timer.timeout.connect(watcher)

            # Execute Slot.on_event_connect()
            return fn(*args, **kwargs)

        # Returns defined wrapper
        return wrapper

    # Decorates Server.history()
    @staticmethod
    def handler(fn):
        # To keep docstring of fn
        @wraps(fn)
        # Define wrapper function
        def wrapper(*args):
            server = args[0]
            code = server.share.get_args('history', 'code')

            # Execute Server.history(*args)
            try:
                fn(*args)

            # RuntimeError can be caused by two cases
            # 1) not monotonic increasing data
            # 2) received wrong data other than requested
            except Exception:
                # Print error message
                print(f'\n[{clock()}] An error at Server.history{args[1:]} with code={code}.\n\n{format_exc()}')
                # Reset variables
                server.share.remove_history(code)
                server.share.remove_args('history')
                server.share.update_single('history', 'error', True)
                # Return to Signal.history()
                server.api.unloop()

        # Returns defined wrapper
        return wrapper



from os import getcwd, makedirs
from os.path import exists, join
from textwrap import dedent
from traceback import format_exc
from warnings import warn

import pandas as pd

from kiwoom import config
from kiwoom.config import history
from kiwoom.config.error import msg
from kiwoom.config.types import MULTI
from kiwoom.core.kiwoom import Kiwoom
from kiwoom.data.share import Share
from kiwoom.data.preps import string
from kiwoom.utils.general import date, name
from kiwoom.utils.manager import Downloader


class Server:
    def __init__(self):
        from kiwoom.core.bot import Bot
        self.bot: Bot = None
        self.api: Kiwoom = None
        self.share: Share = None

    def init(self, bot, api, share):
        self.bot = bot
        self.api = api
        self.share = share

    """
    Default slots to the most basic two events.
        on_event_connect
        on_receive_msg
    """
    # Default event slot for on_event_connect
    def login(self, err_code):
        """
        Default slot for 'on_event_connect'

        When Kiwoom.on_event_connect(...) is called, this method will be automatically called.
        """
        print(f'\n로그인 {msg(err_code)}')
        print(f'\n* 시스템 점검\n  - 월 ~ 토 : 05:05 ~ 05:10\n  - 일 : 04:00 ~ 04:30\n')
        self.api.unloop()

    # Default event slot for on_receive_msg_slot
    def on_receive_msg(self, scr_no, rq_name, tr_code, msg):
        """
        Default slot for 'on_receive_msg'

        Whenever the server sends a message, this method prints depending on below.
        >> Kiwoom.message(True)
        >> Kiwoom.message(False)
        """
        if self.api.msg:
            print(f'\n화면번호: {scr_no}, 요청이름: {rq_name}, TR코드: {tr_code} \n{msg}\n')

    """
    Basic methods
    """
    @Downloader.handler
    def history(self, scr_no, rq_name, tr_code, _, prev_next):
        kwargs = self.share.get_args(name())
        period = history.get_period(tr_code)

        rec = history.get_record_name_for_its_name(tr_code)  # record_name = '종목코드' | '업종코드'
        code = string(self.api.get_comm_data(tr_code, rq_name, 0, rec))

        # Handle trading suspended stock
        if not code:  # code = ''
            code = kwargs['code']

        # Check if wrong data received
        if code != kwargs['code']:
            raise RuntimeError(f"Requested {kwargs['code']}, but the server still sends {code}.")

        # Fetch multi data
        data = {key: list() for key in history.outputs(tr_code, MULTI)}
        cnt = self.api.get_repeat_cnt(tr_code, rq_name)
        for i in range(cnt):
            for key, fn in history.preper(tr_code, MULTI):
                data[key].append(fn(self.api.get_comm_data(tr_code, rq_name, i, key)))

        # Update downloaded data
        for key in data.keys():
            self.share.extend_history(code, key, data[key])

        # If data is more than needed, then stop downloading.
        if 'start' in kwargs:
            col = history.get_datetime_column(period)
            # To check whether it's an empty data.
            if len(data[col]) > 0:
                last = data[col][-1][:len('YYYYMMDD')]
                # Note that data is ordered from newest to oldest
                if date(last) < date(kwargs['start']):
                    prev_next = ''

        # Continue to download
        if prev_next == '2':
            try:
                # Call signal method again, but with prev_next='2'
                bot = self.api.signal('on_receive_tr_data', name())
                bot(code, period=period, prev_next=prev_next)
            except Exception as err:
                args = f"code={code}, period={period}, prev_next={prev_next}"
                self.share.update_single('history', 'error', True)
                print(f"An error at Bot.history({args}).\n\n{format_exc()}")

        # Download done
        else:
            # Sort to chronological order
            df = pd.DataFrame(self.share.get_history(code))[::-1]

            # To make df have datetime index
            col = history.get_datetime_column(period)
            fmt = history.get_datetime_format(period)

            """
                Make time-related column as pandas Datetime index
            """
            # To handle exceptional time and dates
            if not df.empty and history.is_sector(code) and col == '체결시간':
                # To choose exceptional datetime replacer
                edrfec = history.EXCEPTIONAL_DATETIME_REPLACER_FOR_EXCEPTIONAL_CODE
                replacer = edrfec[code] if code in edrfec else history.EXCEPTIONAL_DATETIME_REPLACER

                # Find index of dates that delayed market opening time and inconvertibles in df
                indices = dict()
                exceptions = list()
                start, end = date(df[col].iat[0][:len('YYYYMMDD')]), date(df[col].iat[-1][:len('YYYYMMDD')])
                for ymd, delay in history.EXCEPTIONAL_DATES.items():
                    if start <= date(ymd) <= end:
                        day = df[col].loc[df[col].str.match(ymd)]
                        indices[ymd] = day.index

                        # To save original data
                        for regex, datetime in replacer.items():
                            series = day.loc[day.str.contains(regex, regex=True)]
                            series = series.replace(regex={regex: datetime})
                            series = pd.to_datetime(series, format='%Y%m%d%H%M%S')
                            exceptions.append(series)

                # Replace inconvertibles (888888, 999999) to (16:00:00, 18:00:00)
                df[col].replace(regex=replacer, inplace=True)

                # To make column as pandas datetime series
                df[col] = pd.to_datetime(df[col], format=fmt)

                # Subtract delayed market time as if it pretends to start normally
                for ymd, idx in indices.items():
                    delay = history.EXCEPTIONAL_DATES[ymd]
                    df.loc[idx, col] -= pd.DateOffset(hours=delay)

                # Replace subtracted exceptional times back to original
                for series in exceptions:
                    df.loc[series.index, col] = series

            # col='일자' or including df.empty for both col
            else:
                df[col] = pd.to_datetime(df[col], format=fmt)

            # Finally make datetime column as index
            df.set_index(col, inplace=True)

            """
                Close downloading process
            """
            # To get rid of data preceding 'start'
            if 'start' in kwargs:
                df = df.loc[kwargs['start']:]
            # To get rid of data following 'end'
            if 'end' in kwargs:
                df = df.loc[:kwargs['end']]

            # If server sent mixed data
            if not df.index.is_monotonic_increasing:
                raise RuntimeError(
                    f'Downloaded data is not monotonic increasing. Error at Server.history() with code={code}.'
                )

            # Rename column
            if period == 'tick':
                df.rename(columns={'현재가': '체결가'}, inplace=True)

            # Save data to csv file
            self.history_to_csv(df, code, kwargs['path'], kwargs['merge'], kwargs['warning'])

            # Once common variables are used, delete it
            self.share.remove_args(name())
            self.share.remove_history(code)

            # Mark successfully downloaded
            self.share.update_single(name(), 'complete', True)

            self.api.disconnect_real_data(scr_no)
            self.api.unloop()

    def history_to_csv(self, df, file, path=None, merge=False, warning=True):
        """
        Save historical data of given code at path in .csv format.

        Once the data is saved, it will be removed from the memory.
        When merge is True, data will be merged with existing file.
        Data will be overwritten by default, otherwise.

        :param df: pandas.Dataframe
        :param file: str
        :param path: str
        :param merge : bool
        :param warning: bool
        """
        # In case, path is '' or None
        if not path:
            path = getcwd()

        if not exists(path):
            makedirs(path)

        file = file if file.endswith('.csv') else file + '.csv'
        file = join(path, file)

        if merge:
            # No file to merge with
            if not exists(file):
                # An empty file will be created later
                pass

            # Nothing to be done
            elif df.empty:
                return

            # To merge with existing data
            else:
                col = df.index.name
                if col not in ['일자', '체결시간']:
                    raise ValueError(f"No column matches '일자' or '체결시간'. Merge can't be done.")

                # Existing data
                if 'file' in self.share.single['history']:
                    db = self.share.get_single('history', 'file')
                else:
                    # Read the existing file from disk
                    db = pd.read_csv(
                        file,
                        index_col=[col],
                        parse_dates=[col],
                        encoding=config.ENCODING
                    )
                db.dropna(axis='index', inplace=True)

                if not db.empty:
                    # To check db has more past data, at least the same
                    assert (db.index[0] <= df.index[0]), \
                        f"Existing file starts from {db.index[0]}, while given data from {df.index[0]}."

                    # To check db is chronologically ordered
                    assert db.index.is_monotonic_increasing, \
                        f"The existing file, {file}, is not sorted in chronological order."

                    try:
                        start = db.index.get_loc(df.index[0])
                        # To handle multiple same timestamps
                        if isinstance(start, slice):
                            start = start.start

                        db = db.iloc[:start]
                        df = pd.concat([db, df], axis=0, join='outer', copy=False)

                    except KeyError:
                        err_msg = dedent(
                            f"""
                            Data, '{file}', is forced to be merged but it may not be time-continuous.
                             - The End of the Existing Data : {db.index[-1]}
                             - The Start of Downloaded Data : {df.index[0]}
                            """
                        )
                        # Note that tick data may change depending on downloading time
                        # Kiwoom server may calibrate data after market close (in my opinion)
                        if col == '체결시간':  # tick, min
                            start_date = df.index[0].date()
                            if warning:
                                # The case data may not be time-continuous
                                if db.loc[db.index == start_date].empty:
                                    warn(err_msg)
                            # To slice DB before the date when downloaded data starts from
                            db = db[:start_date]

                        # The case data may not be time-continuous
                        else:  # col == '일자'  # day, week, month, year
                            if warning:
                                warn(err_msg)

                        # Just concatenate if no overlapping period.
                        df = pd.concat([db, df], axis=0, join='outer', copy=False)

        if not df.index.is_monotonic_increasing:
            raise RuntimeError(
                f'Error at Server.history_to_csv(file={file}, ...)/\n'
                + 'File to write is not monotonic increasing with respect to time.'
            )

        # To prevent overwriting
        if not merge and exists(file):
            raise FileExistsError(
                f'Error at Server.history_to_csv(file={file}, ...)/\n'
                + "File already exists. Set merge=True or move the file to prevent from losing data."
            )

        # Finally write to csv file
        df.to_csv(file, encoding=config.ENCODING)



import sys
from inspect import signature

from PyQt5.QtCore import QEventLoop

from kiwoom.config.error import catch_error
from kiwoom.core.connector import Connector
from kiwoom.wrapper.api import API


class Kiwoom(API):
    """
    Main class that can make use of Kiwoom Open API+

    This class wraps all API methods described in KOA studio and Open API+ Manual Guide.
    All methods are wrapped in dynamicCall from PyQt5. Methods and variables names are
    all converted into pythonic way convention, OnReceiveTrData -> on_receive_tr_data.

    Some of extra functions other than pure API are as follows.
    1) Kiwoom.loop() & Kiwoom.unloop()
        These methods are used to prevent from executing codes before data requested
        has not been received yet. Normally, loop can be used in signal methods to stop
        executing further and unloop can be used in slot methods to executing codes that
        are waiting in the signal methods.

    2) Kiwoom.connect(event, signal, slot)
        This method connects signals and slots to one of pre-defined events. Information
        saved in this method is used by decorator @Connector() which wraps the events.

    3) Kiwoom.set_connect_hook(event, param)
        When an event needs multiple slots to connect, depending on specific tasks, set
        a hook(key) to select which slot to map. The hook must be one of the parameters
        in the definition of the event method. Parameters can be found by help built-in
        function or Kiwoom.api_arg_spec(event).

        If hook is set to the given parameter, argument passed into the parameter when
        the event is called, is going to be a key to connect event, signal and slot.

        Convention is that the name of signal and slot that deal with the related task
        is recommended to be the same, so that 'key' is set to be the method name of
        signal and slot by default. See examples on the tutorials link below.
        https://github.com/breadum/kiwoom/blob/main/tutorials/4.%20TR%20Data.py

        Kiwoom.get_connect_hook(), Kiwoom.remove_connect_hook() are also available.

    4) Kiwoom.signal(event, key=None) & Kiwoom.slot(event, key=None)
        If signal and slot are connected to specific event, then these methods return
        the connected signal or slot method respectively. 'key' is needed when hook is
        set by Kiwoom.set_connect_hook().

        When requesting data to server is needed, specifically if more data is available,
        Kiwoom.signal() returns the exact signal method that can request more data.

        When an event is called, Kiwoom.slot() returns the exact slot method that can
        handle data received from the event. This method is used in Connector decorator
        that wraps events to execute connected slot with the event.

    5) @Connector()
        Decorator class that forwards args received from called event into connected slot.
        This class wraps all pre-defined events, and automatically calls connected slots.
    """
    # Class variable just for convenience
    map = Connector.map
    msg = True

    def __init__(self):
        super().__init__()
        self.msg = True
        self._qloop = QEventLoop()

        # To solve the issue that IDE hides error traceback
        def except_hook(cls, exception, traceback):
            sys.__excepthook__(cls, exception, traceback)
        sys.excepthook = except_hook

        # To connect signals and slots
        self._connector = Connector()

        # To set hooks for each event
        self.set_connect_hook('on_receive_tr_data', param='rq_name')
        self.set_connect_hook('on_receive_tr_condition', param='condition_name')
        self.set_connect_hook('on_receive_real_condition', param='condition_name')

        # To connect default slots to basic two events
        # self.connect('on_event_connect', slot=self._slot.on_event_connect)
        # self.connect('on_receive_msg', slot=self._slot.on_receive_msg)

    def loop(self):
        """
        Stop executing codes by running QEventLoop in PyQt5

        If loop is already running, then this does nothing, else freezing. This
        method can be used in signal methods to wait response from the server.
        """
        if not self._qloop.isRunning():
            self._qloop.exec()

    def unloop(self):
        """
        Keep executing codes that are blocked by Kiwoom.loop().

        If loop is not running, then this does nothing, else unfreezing. This
        method can be used in slot methods to execute awaiting codes. If the
        slot methods called, then it means client received response from server.
        There is no need to block codes that are awaiting response, anymore.
        """
        if self._qloop.isRunning():
            self._qloop.exit()

    def login(self):
        """
        Request login to the server by CommConnect API method.

        See '개발가이드 > 로그인 버전처리 > 관련함수 > CommConnect' in KOA Studio.
        """
        self.comm_connect()
        self.loop()

    @classmethod
    def message(cls, bool):
        """
        Turn on/off printing message from Kiwoom.on_receive_msg() event.

        :param bool: bool
            If True, then it prints message else does not.
        """
        cls.msg = bool

    def signal(self, event, key=None):
        """
        Returns signal methods connected to the event.

        If signal and slot are connected to specific event by Kiwoom.connect() method,
        then this method returns the connected signal method. If signal is not connected,
        or wrong key is given, this raises a KeyError.

        'key' is needed when hook is set by Kiwoom.set_connect_hook(). 'key' is set to
        be the name of signal method by default unless another string is set on purpose
        when connecting.

        When requesting data to server is needed, specifically if more data is available,
        Kiwoom.signal() returns the exact signal method that can request more data.

        :param event: str
            One of the pre-defined event names in string. See kiwoom.config.events.
        :param key: str, optional
            If hook is set by Kiwoom.set_connect_hook() method and signal is connected
            by Kiwoom.connect(), then key is needed. 'key' is set to be name of the
            signal method by default unless another 'key' is given when connecting.
        :return: method
            Signal method connected to the given event. If wrong event, returns None.
        """
        return self._connector.signal(event, key)

    def slot(self, event, key=None):
        """
        Returns slot methods connected to the event.

        If signal and slot are connected to specific event by Kiwoom.connect() method,
        then this method returns the connected slot method. If slot is not connected,
        or wrong key is given, this raises a KeyError.

        'key' is needed when hook is set by Kiwoom.set_connect_hook(). 'key' is set to
        be the name of slot method by default unless another string is set on purpose
        when connecting.

        When an event is called, Kiwoom.slot() returns the exact slot method that can
        handle data received from the event. This method is used in Connector decorator
        that wraps events to execute connected slot with the event.

        :param event: str
            One of the pre-defined event names in string. See kiwoom.config.events.
        :param key: str, optional
            If hook is set by Kiwoom.set_connect_hook() method and slot is connected
            by Kiwoom.connect(), then key is needed. 'key' is set to be name of the
            slot method by default unless another 'key' is given when connecting.
        :return: method or None
            Slot method connected to the given event. If wrong event, returns None.
        """
        return self._connector.slot(event, key)

    def connect(self, event, signal=None, slot=None, key=None):
        """
        Connects signals and slots to one of pre-defined events.

        Information saved in this method is used by decorator @Connector() which wraps
        the events and automatically calls the right slot connected to the events. In
        addition to the decorator, Kiwoom.signal(event, key) and Kiwoom.slot(event, key)
        returns the one connected to the event.

        1) If no hook is set on the event, then the connected signal/slot can be retrieved
           by Kiwoom.signal(event) and Kiwoom.slot(event). There is no need to use key.

        2) If hook is set by Kiwoom.set_connect_hook() on the event, in which case there
           needs multiple slots to connect on one event, then connection requires a key
           which is to be the name of signal/slot methods by default.

           The convention to utilizing this module recommends to define the name of related
           signal and slot to be the same. Then it becomes easier to manage and develop codes.

           Use 'key' arg only when there is a special need. The connected signal/slot can be
           retrieved by Kiwoom.signal(event, key='name') and Kiwoom.slot(event, key='name').
           Here 'name' can be a method name or special 'key' used in this method.

        This method checks whether or not given signal/slot can be called without any
        problem. If given method is not bounded to an instance, method should be static
        or lambda function. This is because normally 'self' argument is needed to call
        methods, therefore method must be bounded to an instance unless given method is
        a function.

        Please see tutorials example on the link below.
        https://github.com/breadum/kiwoom/blob/main/tutorials/4.%20TR%20Data.py

        :param event: str
            One of the pre-defined event names in string. See kiwoom.config.events.
        :param signal: method, optional
            A method that requests to the server
        :param slot: method, optional
            A method that reacts the server's response
        :param key: str, optional
            Key is needed only if hook is set by Kiwoom.set_connect_hook() method.
            Key is set to be name of the given signal and/or slot method by default.
            If key is given other than method name, the connected signal can be
            retrieved by Kiwoom.siganl(event, key) and slot by Kiwoom.slot(event, key)
        """
        self._connector.connect(event, signal=signal, slot=slot, key=key)

    def set_connect_hook(self, event, param):
        """
        Set parameter defined in event as a hook to find the right slot when event is called.

        When an event needs multiple slots to connect, depending on specific tasks, set
        a hook(key) to select which slot to map. The hook must be one of the parameters
        in the definition of the event method. Parameters can be found by help built-in
        function or Kiwoom.api_arg_spec(event). This raises a KeyError if given param is
        not defined in event method.

        If hook is set to the given parameter, argument passed into the parameter when
        the event is called, is going to be a key to connect event, signal and slot.

        The Convention is that the names of signal and slot that deal with the related task
        are recommended to be the same, so that 'key' is set to be the method name of
        signal and slot by default. See examples on the tutorials link below.
        https://github.com/breadum/kiwoom/blob/main/tutorials/4.%20TR%20Data.py

        :param event: str
            One of the pre-defined event names in string. See kiwoom.config.events.
        :param param: str
            Parameter name defined in given event. To see all parameters to event,
            use Kiwoom.api_arg_spec(event) method or help(...) built-in function.
        """
        self._connector.set_connect_hook(event, param)

    def get_connect_hook(self, event):
        """
        Returns a hook (i.e. name of parameter) set in given event.

        :param event: str
            One of the pre-defined event names in string. See kiwoom.config.events.
        :return: str or None
            If exists, returns hook in string else None. If not a valid event is given,
            this returns None.
        """
        return self._connector.get_connect_hook(event)

    def remove_connect_hook(self, event):
        """
        Remove hook which is set in given event if exists.

        This method removes all information of signals and slots connected to given
        event as well as hook. If hook of given event does not exist, this raises
        a KeyError.

        :param event: str
            One of the pre-defined event names in string. See kiwoom.config.events.
        """
        self._connector.remove_connect_hook(event)

    def get_hook_index(self, event):
        """
        Returns index of hook in method arguments.

        :param event: str
            One of the pre-defined event names in string. See kiwoom.config.events.
        :return: int
        """
        return self._connector.get_hook_index(event)

    @staticmethod
    def api_arg_spec(fn):
        """
        Returns a string list of parameters to given API function

        :param fn: str
            Name of API function to get list of parameters.
        :return: list
            Parameters of given API function in list of strings.
        """
        return signature(getattr(API, fn)).parameters

    """
    Event Handlers (8)
        on_event_connect
        on_receive_msg
        on_receive_tr_data
        on_receive_real_data
        on_receive_chejan_data
        on_receive_condition_ver
        on_receive_tr_condition
        on_receive_real_condition
    """
    @map
    def on_event_connect(self, err_code):
        pass

    @map
    def on_receive_msg(self, scr_no, rq_name, tr_code, msg):
        pass

    @map
    def on_receive_tr_data(self, scr_no, rq_name, tr_code, record_name, prev_next):
        pass

    @map
    def on_receive_real_data(self, code, real_type, real_data):
        pass

    @map
    def on_receive_chejan_data(self, gubun, item_cnt, fid_list):
        pass

    @map
    def on_receive_condition_ver(self, ret, msg):
        pass

    @map
    def on_receive_tr_condition(self, scr_no, code_list, condition_name, index, next):
        pass

    @map
    def on_receive_real_condition(self, code, type, condition_name, condition_index):
        pass

    """
    Methods that return error codes.
        comm_rq_data
        comm_kw_rq_data
        send_order
        send_order_fo
        send_order_credit
        set_real_reg
    """
    @catch_error
    def comm_rq_data(self, rq_name, tr_code, prev_next, scr_no):
        return super().comm_rq_data(rq_name, tr_code, prev_next, scr_no)

    @catch_error
    def comm_kw_rq_data(self, arr_code, next, code_cnt, type_flag, rq_name, scr_no):
        return super().comm_kw_rq_data(arr_code, next, code_cnt, type_flag, rq_name, scr_no)

    @catch_error
    def send_order(self, rq_name, scr_no, acc_no, ord_type, code, qty, price, hoga_gb, org_order_no):
        return super().send_order(rq_name, scr_no, acc_no, ord_type, code, qty, price, hoga_gb, org_order_no)

    @catch_error
    def send_order_fo(self, rq_name, scr_no, acc_no, code, ord_kind, sl_by_tp, ord_tp, qty, price, org_ord_no):
        return super().send_order_fo(rq_name, scr_no, acc_no, code, ord_kind, sl_by_tp, ord_tp, qty, price, org_ord_no)

    @catch_error
    def send_order_credit(
            self,
            rq_name,
            scr_no,
            acc_no,
            order_type,
            code,
            qty,
            price,
            hoga_gb,
            credit_gb,
            loan_date,
            org_order_no
    ):
        return super().send_order_credit(
            rq_name,
            scr_no,
            acc_no,
            order_type,
            code,
            qty,
            price,
            hoga_gb,
            credit_gb,
            loan_date,
            org_order_no
        )

    @catch_error
    def set_real_reg(self, scr_no, code_list, fid_list, opt_type):
        return super().set_real_reg(scr_no, code_list, fid_list, opt_type)



from collections import defaultdict
from random import randint


MAX_SCREEN_COUNT = 200
MAX_STOCK_PER_SCREEN = 90


class Screen:
    def __init__(self):
        self.config = {
            'opt10079': '4000',  # Stock tick
            'opt10080': '4080',  # Stock min
            'opt10081': '4160',  # Stock day
            'opt10082': '4240',  # Stock week
            'opt10083': '4320',  # Stock month
            'opt10094': '4400',  # Stock year
            'opt20004': '4500',  # Sector tick
            'opt20005': '4580',  # Sector min
            'opt20006': '4660',  # Sector day
            'opt20007': '4740',  # Sector week
            'opt20008': '4820',  # Sector month
            'opt20019': '4900'  # Sector year
        }
        self._used = set()  # assigned screen numbers to TR code
        self._alloc = defaultdict(dict)  # assigned screen number to stock
        self._count = defaultdict(lambda: 0)  # assigned number of stocks to screen number

    def __call__(self, tr_code):
        if tr_code in self:
            self._used.add(self.config[tr_code])
            return self.config[tr_code]

        if len(self._used) >= MAX_SCREEN_COUNT:
            raise RuntimeError(f'The number of screen exceeds maximum limit {MAX_SCREEN_COUNT}.')

        # Assign new random screen number to tr code
        scr_no = str(randint(0, 9999)).zfill(4)
        while scr_no in self:
            scr_no = str(randint(0, 9999)).zfill(4)
        self._used.add(scr_no)
        self.config[tr_code] = scr_no
        return scr_no

    def __contains__(self, key):
        return (key in self.config) or (key in self._used)

    def __delitem__(self, tr_code):
        self._used.remove(tr_code)
        del self.config[tr_code]

    def update(self, tr_code, scr_no):
        self.config[tr_code] = scr_no

    def alloc(self, tr_code, code):
        if code in self._alloc[tr_code]:
            return self._alloc[tr_code][code]

        scr_no = self(tr_code)
        # To check the number of allocated stock/sector exceeds maximum elements per screen
        if self._count[scr_no] >= MAX_STOCK_PER_SCREEN:
            # Try out for the next screen number first
            scr_no = str(int(scr_no.lstrip('0')) + 1).zfill(4)
            # In case next number is already in use
            while scr_no in self:
                rnd = randint(1, 100)
                scr_no = (int(scr_no.lstrip('0')) + rnd) % 10000
                scr_no = str(scr_no).zfill(4)
            self.update(tr_code, scr_no)

        self._alloc[tr_code][code] = scr_no
        self._count[scr_no] += 1
        return self(tr_code)



from setuptools import setup, find_packages


# Project Information
VERSION = "1.3.5"
DISTNAME = "kiwoom"
DESCRIPTION = "Simple Python Wrapper for Kiwoom Open API+"
LONG_DESCRIPTION = open("README.md", encoding="utf-8").read()
LONG_DESCRIPTION_CONTENT_TYPE = 'text/markdown'
URL = "https://github.com/breadum/kiwoom"
DOWNLOAD_URL = "https://pypi.org/project/kiwoom"
PROJECT_URLS = {
    "Git": "https://github.com/breadum/kiwoom",
    "Tutorials": "https://github.com/breadum/kiwoom/tree/main/tutorials"
}

# Author Information
AUTHOR = "Breadum"
EMAIL = "breadum.kr@gmail.com"

# Miscellaneous
LICENSE = "MIT License"
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "License :: OSI Approved :: MIT License",
    "Operating System :: Microsoft :: Windows",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Topic :: Office/Business :: Financial :: Investment"
]
KEYWORDS =[
    "Kiwoom",
    "Heroes",
    "Open API+",
    "키움",
    "영웅문",
    "System Trading",
    "Algorithmic Trading"
]


def setup_package():
    kwargs = {
        "install_requires": [
            'tabulate >= 0.9.0',
            'PyQt5 >= 5.15.9'
        ],
        "setup_requires": [],
        'zip_safe': False
    }

    setup(
        name=DISTNAME,
        version=VERSION,
        author=AUTHOR,
        author_email=EMAIL,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
        url=URL,
        download_url=DOWNLOAD_URL,
        project_urls=PROJECT_URLS,
        packages=find_packages(),
        python_requires=">=3.6",
        license=LICENSE,
        classifiers=CLASSIFIERS,
        keywords=KEYWORDS,
        **kwargs
    )


if __name__ == '__main__':
    setup_package()
